from abc import ABC, abstractmethod
from pydantic import BaseModel, field_validator
from typing import List, Optional, Any, Dict, Union

from adala.utils.internal_data import InternalDataFrame


class Dataset(BaseModel, ABC):
    """
    Abstract base class representing a dataset.

    Provides methods to interact with and obtain information about datasets. 
    Concrete implementations should provide functionality for batch iteration, 
    getting dataset size, and displaying dataset information.
    """

    @abstractmethod
    def batch_iterator(self, batch_size: int = 100) -> InternalDataFrame:
        """
        Yields batches of data records from the dataset.
        
        Args:
            batch_size (int, optional): Size of each batch to be yielded. Defaults to 100.
            
        Returns:
            InternalDataFrame: A data frame containing a batch of records.
        """

    @abstractmethod
    def __len__(self) -> int:
        """
        Provides the number of records in the dataset.
        
        Returns:
            int: Total number of records in the dataset.
        """

    @abstractmethod
    def info(self) -> None:
        """
        Displays information about the dataset.
        """


class BlankDataset(Dataset):
    """
    Represents an empty dataset with no records.

    This class can be used in situations where a dataset is required, 
    but no actual data is available or needed. 
    All methods return defaults representing an empty state.
    """

    def batch_iterator(self, batch_size: int = 100) -> InternalDataFrame:
        """
        Yields an empty data frame as there are no records in a blank dataset.
        
        Args:
            batch_size (int, optional): This argument is ignored for BlankDataset. Defaults to 100.
            
        Returns:
            InternalDataFrame: An empty data frame.
        """
        
        return InternalDataFrame()

    def __len__(self) -> int:
        """
        Provides the number of records in the blank dataset (which is always 0).
        
        Returns:
            int: Total number of records in the dataset (0 for BlankDataset).
        """
        
        return 0

    def info(self) -> None:
        """
        Displays information about the blank dataset.
        """
        
        print('Blank dataset')



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any, Optional, TYPE_CHECKING

from pydantic import BaseModel
from adala.datasets.base import Dataset, InternalDataFrame
from rich import print

if TYPE_CHECKING:
    from adala.skills.skillset import SkillSet


class ShortTermMemory(BaseModel):
    """
    Base class for short term memory storage
    """
    dataset: Dataset = None
    predictions: InternalDataFrame = None
    evaluations: InternalDataFrame = None
    ground_truth_column_name: str = None
    match_column_name: str = None
    errors: InternalDataFrame = None
    accuracy: float = None
    initial_instructions: str = None
    updated_instructions: str = None

    class Config:
        arbitrary_types_allowed = True

    def reset(self):
        self.predictions = None
        self.evaluations = None
        self.errors = None
        self.accuracy = None
        self.initial_instructions = None
        self.updated_instructions = None

    def __rich__(self):
        text = '[bold blue]Agent Experience:[/bold blue]\n\n'
        if self.predictions is not None:
            text += f'\n[bold]Predictions[/bold]\n{self.predictions}'
        if self.evaluations is not None:
            text += f'\n[bold]Evaluations[/bold]\n{self.evaluations}'
        if self.errors is not None:
            text += f'\n[bold]Errors[/bold]\n{self.errors}'
        if self.accuracy is not None:
            text += f'\n[bold]Accuracy[/bold]\n{self.accuracy}'
        if self.initial_instructions is not None:
            text += f'\n[bold]Initial Instructions[/bold]\n{self.initial_instructions}'
        if self.updated_instructions is not None:
            text += f'\n[bold]Updated Instructions[/bold]\n{self.updated_instructions}'
        return text

    def display(self):
        print(self)


class LongTermMemory(BaseModel, ABC):

    """
    Base class for long-term memories.
    Long-term memories are used to store acquired knowledge and can be shared between agents.
    """

    @abstractmethod
    def remember(self, experience: ShortTermMemory, skills: SkillSet):
        """
        Base method for remembering experiences in long term memory.
        """

    @abstractmethod
    def retrieve(self, observations: ShortTermMemory) -> ShortTermMemory:
        """
        Base method for retrieving past experiences from long term memory, based on current observations
        """



from ..base import LLMSkill


class TextGenerationSkill(LLMSkill):
    """
    Skill specialized for generating text based on the provided input.

    This involves tasks where the LLM is expected to produce creative, coherent, and contextually 
    relevant textual content based on the given input.

    Attributes:
        instructions (str): Instruction to guide the LLM in text generation.
    """
    
    instructions: str = 'Generate text based on the provided input.'



from .base import TextGenerationSkill


class QuestionAnsweringSkill(TextGenerationSkill):
    """
    Skill specialized for answering questions based on the provided input.

    Inherits from the TextGenerationSkill and focuses on generating answers to the questions 
    posed in the input. The class customizes the instructions, input, and output templates 
    specifically for question-answering tasks.

    Attributes:
        instructions (str): Instruction to guide the LLM in answering the question.
        input_template (str): Format in which the question is presented to the LLM.
        output_template (str): Expected format of the LLM's answer.
        prediction_field (str): Field name for the generated answer.
    """
    
    instructions: str = 'Answer the question.'
    input_template: str = "Question: {{{{{input}}}}}"
    output_template: str = "Answer: {{gen 'answer'}}"
    prediction_field: str = 'answer'




from pydantic import BaseModel, Field, SkipValidation, field_validator, model_validator, ValidationError
from abc import ABC, abstractmethod
from typing import Any, Optional, List, Dict, Union
from adala.environments.base import Environment, BasicEnvironment
from adala.datasets import Dataset, DataFrameDataset
from adala.runtimes.base import Runtime, LLMRuntime, LLMRuntimeType, LLMRuntimeModelType
from adala.runtimes.openai import OpenAIRuntime
from adala.memories.base import ShortTermMemory, LongTermMemory
from adala.skills.base import BaseSkill
from adala.skills.skillset import SkillSet, LinearSkillSet
from adala.utils.logs import print_dataframe, print_text, print_error
from adala.utils.internal_data import InternalDataFrame


class Agent(BaseModel, ABC):
    """
    Represents a customizable agent that can interact with environments, 
    employ skills, and leverage memory and runtimes.

    Attributes:
        environment (Union[Dataset, Environment]): The environment with which the agent interacts.
        skills (Union[SkillSet, BaseSkill, List[BaseSkill], Dict[str, BaseSkill]]): The skills possessed by the agent.
        memory (LongTermMemory, optional): The agent's long-term memory. Defaults to None.
        runtimes (Dict[str, Runtime], optional): The runtimes available to the agent. Defaults to predefined runtimes.
        default_runtime (str): The default runtime used by the agent. Defaults to 'openai'.
    """
    
    environment: Union[InternalDataFrame, Dataset, Environment] = Field(default_factory=DataFrameDataset)
    skills: Union[SkillSet, BaseSkill, List[BaseSkill], Dict[str, BaseSkill]]

    memory: LongTermMemory = Field(default=None)
    runtimes: Optional[Dict[str, Runtime]] = Field(
        default_factory=lambda: {
            'openai': OpenAIRuntime(model='gpt-3.5-turbo-instruct'),
            # 'llama2': LLMRuntime(
            #     llm_runtime_type=LLMRuntimeModelType.Transformers,
            #     llm_params={
            #         'model': 'meta-llama/Llama-2-7b',
            #         'device': 'cuda:0',
            #     }
            # )
        }
    )
    teacher_runtimes: Optional[Dict[str, Runtime]] = Field(
        default_factory=lambda: {
            'openai-gpt3': OpenAIRuntime(model='gpt-3.5-turbo'),
            'openai-gpt4': OpenAIRuntime(model='gpt-4')
        }
    )
    default_runtime: str = 'openai'
    default_teacher_runtime: str = 'openai-gpt3'

    class Config:
        arbitrary_types_allowed = True

    def __rich__(self):
        """
        Returns a colorized and formatted representation of the Agent instance.

        Returns:
            str: A rich-formatted representation of the agent.
        """
        
        skill_names = ", ".join([skill.name for skill in self.skills.skills.values()])
        runtime_names = ", ".join(self.runtimes.keys())
        
        return (
            f"[bold blue]Agent Instance[/bold blue]\n\n"
            f"Environment: {self.environment.__class__.__name__}\n"
            f"Skills: {skill_names}\n"
            f"Runtimes: {runtime_names}\n"
            f"Default Runtime: {self.default_runtime}\n"
            f"Default Teacher Runtime: {self.default_teacher_runtime}"
        )

    @field_validator('environment')
    def environment_validator(cls, v):
        """
        Validates and possibly transforms the environment attribute.

        Args:
            v (Union[Dataset, Environment]): The environment value to validate.

        Returns:
            Environment: The validated environment.
        """
        if isinstance(v, InternalDataFrame):
            v = DataFrameDataset(df=v)
        if isinstance(v, Dataset):
            v = BasicEnvironment(dataset=v)
        return v

    @field_validator('skills')
    def skills_validator(cls, v):
        """
        Validates and possibly transforms the skills attribute.

        Args:
            v (Union[SkillSet, BaseSkill, List[BaseSkill], Dict[str, BaseSkill]]): The skills value to validate.

        Returns:
            SkillSet: The validated set of skills.
        """
        
        if isinstance(v, SkillSet):
            pass
        elif isinstance(v, BaseSkill):
            v = LinearSkillSet(skills={'skill_0': v})
        elif isinstance(v, list):
            v = LinearSkillSet(skills={f'skill_{i}': skill for i, skill in enumerate(v)})
        elif isinstance(v, dict):
            v = LinearSkillSet(skills=v)
        return v

    @model_validator(mode='after')
    def verify_input_parameters(self):
        def _raise_default_runtime_error(val, runtime, runtimes, default_value):
            print_error(f"The Agent.{runtime} is set to {val}, "
                        f"but this runtime is not available in the list: {list(runtimes)}. "
                        f"Please choose one of the available runtimes and initialize the agent again, for example:\n\n"
                        f"agent = Agent(..., {runtime}='{default_value}')\n\n"
                        f"Make sure the default runtime is available in the list of runtimes. For example:\n\n"
                        f"agent = Agent(..., runtimes={{'{default_value}': OpenAIRuntime(model='gpt-4')}})\n\n")
            raise ValueError(f"default runtime {val} not found in provided runtimes.")

        if self.default_runtime not in self.runtimes:
            _raise_default_runtime_error(self.default_runtime, 'default_runtime', self.runtimes, 'openai')
        if self.default_teacher_runtime not in self.teacher_runtimes:
            _raise_default_runtime_error(self.default_teacher_runtime, 'default_teacher_runtime', self.teacher_runtimes, 'openai-gpt4')
        return self

    def get_runtime(self, runtime: Optional[str] = None) -> Runtime:
        """
        Retrieves the specified runtime or the default runtime if none is specified.

        Args:
            runtime (str, optional): The name of the runtime to retrieve. Defaults to None.

        Returns:
            Runtime: The requested runtime.

        Raises:
            ValueError: If the specified runtime is not found.
        """
        
        if runtime is None:
            runtime = self.default_runtime
        if runtime not in self.runtimes:
            raise ValueError(f'Runtime "{runtime}" not found.')
        return self.runtimes[runtime]

    def get_teacher_runtime(self, runtime: Optional[str] = None) -> Runtime:
        """
        Retrieves the specified teacher runtime or the default runtime if none is specified.

        Args:
            runtime (str, optional): The name of the runtime to retrieve. Defaults to None.

        Returns:
            Runtime: The requested runtime.

        Raises:
            ValueError: If the specified runtime is not found.
        """

        if runtime is None:
            runtime = self.default_teacher_runtime
        if runtime not in self.teacher_runtimes:
            raise ValueError(f'Teacher Runtime "{runtime}" not found.')
        return self.teacher_runtimes[runtime]

    def apply_skills(
        self,
        dataset: Union[Dataset, InternalDataFrame],
        runtime: Optional[Union[str, Runtime]] = None,
        experience: Optional[ShortTermMemory] = None,
    ) -> ShortTermMemory:
        """
        Applies the agent's skills to a given dataset using the specified runtime.

        Args:
            dataset (Dataset): The dataset to apply skills on.
            runtime (str, optional): The runtime to use. Defaults to None.
            experience (ShortTermMemory, optional): The agent's short-term memory. Defaults to None.

        Returns:
            ShortTermMemory: The short-term memory resulting from the application of skills.
        """
        runtime = runtime or self.default_runtime
        if isinstance(dataset, InternalDataFrame):
            dataset = DataFrameDataset(df=dataset)
        if isinstance(runtime, str):
            runtime = self.get_runtime(runtime=runtime)
        return self.skills.apply(dataset=dataset, runtime=runtime, experience=experience)

    def learn(
        self,
        learning_iterations: int = 3,
        accuracy_threshold: float = 0.9,
        update_skills: bool = True,
        update_memory: bool = True,
        request_environment_feedback: bool = True,
        experience: Optional[ShortTermMemory] = None,
        runtime: Optional[str] = None,
    ) -> ShortTermMemory:
        """
        Enables the agent to learn and improve its skills based on interactions with its environment.

        Args:
            learning_iterations (int, optional): The number of iterations for learning. Defaults to 3.
            accuracy_threshold (float, optional): The desired accuracy threshold to reach. Defaults to 0.9.
            update_skills (bool, optional): Flag to determine if skills should be updated after learning. Defaults to True.
            update_memory (bool, optional): Flag to determine if memory should be updated after learning. Defaults to True.
            request_environment_feedback (bool, optional): Flag to determine if feedback should be requested from the environment. Defaults to True.
            experience (ShortTermMemory, optional): Initial experience for the learning process. Defaults to None.
            runtime (str, optional): The runtime to be used for the learning process. Defaults to None.

        Returns:
            ShortTermMemory: The short-term memory after the learning process.
        """
        
        runtime = self.get_runtime(runtime=runtime)
        # TODO: support teacher runtime input, not default
        teacher_runtime = self.get_teacher_runtime(runtime=self.default_teacher_runtime)

        skills = self.skills.model_copy(deep=True)
        dataset = self.environment.as_dataset()

        # Apply agent skills to dataset and get experience with predictions
        experience = self.apply_skills(dataset=dataset, runtime=runtime, experience=experience)

        # Agent select one skill to improve
        learned_skill = skills.select_skill_to_improve(experience)

        # Request feedback from environment is necessary
        if request_environment_feedback:
            self.environment.request_feedback(learned_skill, experience)

        for iteration in range(learning_iterations):
            print_text(f'\n\n=> Iteration #{iteration}: Comparing to ground truth, analyzing and improving ...')

            # 1. EVALUATION PHASE: Compare predictions to ground truth
            experience = self.environment.compare_to_ground_truth(learned_skill, experience)
            print_text(f'Comparing predictions to ground truth data ...')
            print_dataframe(experience.evaluations)

            # 2. ANALYSIS PHASE: Analyze evaluation experience, optionally use long term memory
            print_text(f'Analyze evaluation experience ...')
            experience = learned_skill.analyze(
                experience=experience,
                student_runtime=runtime,
                teacher_runtime=teacher_runtime,
                memory=self.memory
            )
            print_text(f'Number of errors: {len(experience.errors)}')

            print_text(f'Accuracy = {experience.accuracy*100:0.2f}%', style='bold red')
            if experience.accuracy >= accuracy_threshold:
                print_text(f'Accuracy threshold reached ({experience.accuracy} >= {accuracy_threshold})')
                break

            # 3. IMPROVEMENT PHASE: Improve skills based on analysis
            print_text(f"Improve \"{learned_skill.name}\" skill based on analysis ...")
            experience = learned_skill.improve(
                experience=experience,
                runtime=teacher_runtime,
                update_instructions=True
            )
            print_text(f'Updated instructions for skill "{learned_skill.name}":\n')
            print_text(learned_skill.instructions, style='bold green')

            # 4. RE-APPLY PHASE: Re-apply skills to dataset
            print_text(f"Re-apply {learned_skill.name} skill to dataset ...")
            experience = learned_skill.apply(dataset, runtime, experience=experience)

        # Update skills and memory based on experience
        if update_skills:
            self.skills = skills

        if self.memory and update_memory:
            self.memory.remember(experience, self.skills)

        print_text('Train is done!')
        return experience



import openai
import pandas as pd
import re

from pydantic import BaseModel
from typing import List, Optional, Any, Dict, Tuple
from abc import ABC, abstractmethod
from pydantic import Field, model_validator

from typing import Optional
from adala.runtimes.base import LLMRuntime
from adala.datasets.base import Dataset
from adala.runtimes.base import Runtime
from adala.memories.base import ShortTermMemory, LongTermMemory
from adala.utils.internal_data import InternalDataFrame, InternalDataFrameConcat
from adala.utils.logs import print_error


class BaseSkill(BaseModel, ABC):
    """
    A foundational abstract class representing a skill. This class sets the foundation 
    for all skills and provides common attributes and methods for skill-based operations.
    """
    name: str = Field(
        title='Skill name',
        description='Unique name of the skill',
        default='',
        examples=['labeling', 'classification', 'text-generation']
    )
    instructions: str = Field(
        title='Skill instructions',
        description='Instructs agent what to do with the input data. '
                    'Can use templating to refer to input fields.',
        default='',
        examples=['Label the input text with the following labels: {{labels}}']
    )
    description: Optional[str] = Field(
        default='',
        title='Skill description',
        description='Description of the skill. Can be used to retrieve skill from the library.',
        examples=['The skill to perform sentiment analysis on the input text.']
    )
    input_template: Optional[str] = Field(
        title='Input template',
        description='Template for the input data. '
                    'Can use templating to refer to input parameters and perform data transformations.',
        default="Input: {{{{{input}}}}}",
        examples=["Text: {{{{{input}}}}}, Date: {{{{date_column}}}}, Sentiment: {{{{gen 'sentiment'}}}}"]
    )
    input_data_field: Optional[str] = Field(
        title='Input data field',
        description='Input data field name that will be used to match input data.',
        examples=['text'],
        # TODO: either make it required, or `input_template` required
        default=None
    )
    output_template: Optional[str] = Field(
        title='Output template',
        description='Template for the output data. '
                    'Can use templating to refer to input parameters and perform data transformations. '
                    'Should contain at least one field matching `validation_fields`.',
        default="Output: {{gen 'predictions'}}",
        examples=["Output: {{select 'predictions' options=labels logprobs='score'}}"]
    )
    prediction_field: Optional[str] = Field(
        title='Prediction field',
        description='Prediction field name that will be used to match ground truth labels.'
                    'Should match at least one output field in `output_template`, e.g. \'predictions\'',
        examples=['predictions'],
        default='predictions'
    )

    @model_validator(mode='after')
    def validate_inputs(self):
        """
        Validates the input_template, updating it if necessary.
        
        Returns:
            BaseSkill: Updated instance of the BaseSkill class.
        """
        if '{{{{{input}}}}}' in self.input_template:
            if self.input_data_field is None:
                print_error(f'You provided skill "{self.name}" with input template:\n\n'
                            f'{self.__class__.__name__}.input_template = "{self.input_template}"\n\n'
                            'that contains "{{{{{input}}}}}" placeholder. (yes... 5 curly braces!) \n\n'
                            'In this case, you have to provide skill with `skill.input_data_field` to match the input data.'
                            f'\nFor example, if your input data stored in `"text"` column, '
                            f'you can set\n\nskill = {self.__class__.__name__}(..., input_data_field="text")')
                raise ValueError(f'`input_data_field` is not provided for skill {self.name}')
            self.input_template = self.input_template.format(input=self.input_data_field)
        return self

    def __call__(self, input: InternalDataFrame, runtime: Runtime, dataset: Dataset) -> InternalDataFrame:
        """Calls the runtime to process a batch of inputs. Input and
        output shapes can be varying, and it should also take care of
        data types validation

        Args:
            input (InternalDataFrame): Input data in the form of an InternalDataFrame.
            runtime (Runtime): The runtime instance to be used for processing.
            dataset (Dataset): The dataset containing the data to be processed.
        
        Returns:
            InternalDataFrame: Concatenated dataframe with the original input and the predictions from the runtime.

        """

        # get user defined dataset input fields

        runtime_predictions = runtime.process_batch(
            batch=input,
            input_template=self.input_template,
            output_template=self.output_template,
            instructions=self.instructions,
            extra_fields=self._get_extra_fields()
        )
        return InternalDataFrameConcat((input, runtime_predictions), axis=1)

    def _get_extra_fields(self):
        """
        Retrieves fields that are not categorized as system fields.
        
        Returns:
            dict: A dictionary containing fields that are not system fields.
        """
        
        # TODO: more robust way to exclude system fields
        system_fields = {
            'name', 'description', 'input_template', 'output_template', 'instructions', 'validation_fields'}
        extra_fields = self.model_dump(exclude=system_fields)
        return extra_fields

    @abstractmethod
    def apply(
        self, dataset: Dataset,
        runtime: Runtime,
        experience: ShortTermMemory
    ) -> ShortTermMemory:
        """
        Applies the skill to a dataset and returns the results.
        
        Args:
            dataset (Dataset): The dataset on which the skill is to be applied.
            runtime (Runtime): The runtime instance to be used for processing.
            experience (ShortTermMemory): Previous experiences or results.
        
        Returns:
            ShortTermMemory: The updated experience after applying the skill.
        """        

    @abstractmethod
    def analyze(
        self, experience: ShortTermMemory,
        student_runtime: Runtime,
        teacher_runtime: Optional[Runtime] = None,
        memory: Optional[LongTermMemory] = None,
    ) -> ShortTermMemory:
        """
        Analyzes the results to derive new experiences.
        
        Args:
            experience (ShortTermMemory): The current experience.
            student_runtime (Runtime): The student runtime instance. Defaults to None.
            teacher_runtime (Runtime, optional): The teacher runtime instance. Defaults to None.
            memory (LongTermMemory, optional): Previous long term memories. Defaults to None.

        Returns:
            ShortTermMemory: The updated experience after analysis.
        """

    @abstractmethod
    def improve(
        self,
        experience: ShortTermMemory,
        runtime: Runtime,
        update_instructions: bool = True,
    ) -> ShortTermMemory:
        """
        Refines the current state of the skill based on its experiences.
        
        Args:
            experience (ShortTermMemory): The current experience.
            runtime (Runtime): The runtime instance to be used for processing.
            update_instructions (bool, optional): Flag to decide if instructions should be updated. Defaults to True.
        
        Returns:
            ShortTermMemory: The updated experience after improvements.
        """


class LLMSkill(BaseSkill):
    """
    A skill specialized for Language Models (LLM). Inherits from the BaseSkill 
    class and provides specific implementations for handling LLM predictions based 
    on given instructions.
    """

    def apply(
        self,
        dataset: Dataset,
        runtime: LLMRuntime,
        experience: ShortTermMemory
    ) -> ShortTermMemory:
        """
        Applies the LLM skill on a dataset and returns the results.
        
        Args:
            dataset (Dataset): The dataset on which the skill is to be applied.
            runtime (LLMRuntime): The runtime instance to be used for processing.
            experience (ShortTermMemory): Previous experiences or results.
        
        Returns:
            ShortTermMemory: The updated experience after applying the skill.
        """
        
        experience = experience.model_copy()

        predictions = []

        for batch in dataset.batch_iterator():
            runtime_predictions = self(batch, runtime, dataset)
            predictions.append(runtime_predictions)

        if not predictions:
            predictions = InternalDataFrame()
        else:
            predictions = InternalDataFrameConcat(predictions, copy=False)
            predictions.rename(columns={self.prediction_field: self.name}, inplace=True)

        # append predictions to existing experience, to chain skills
        # TODO: implement predictions chaining
        experience.predictions = predictions
        # if experience.predictions is None:
        #     experience.predictions = predictions
        # else:
        #     experience.predictions = InternalDataFrameConcat([
        #         experience.predictions.drop(columns=[col for col in experience.predictions.columns if col in predictions.columns]),
        #         predictions
        #     ], axis=1)
        #     raise NotImplementedError

        return experience

    def analyze(
        self, experience: ShortTermMemory,
        student_runtime: Runtime,
        teacher_runtime: Optional[Runtime] = None,
        memory: Optional[LongTermMemory] = None
    ) -> ShortTermMemory:
        """
        Analyzes the results to identify any discrepancies and returns the observed experience.
        
        Args:
            experience (ShortTermMemory): The current experience.
            student_runtime (Runtime): The student runtime instance. Defaults to None.
            teacher_runtime (Runtime, optional): The teacher runtime instance. Defaults to None.
            memory (LongTermMemory, optional): Previous long term memories. Defaults to None.

        Returns:
            ShortTermMemory: The updated experience after analysis.
        """
        
        experience = experience.model_copy()

        # TODO: can be multiple prediction validation fields
        match = experience.match_column_name
        errors = experience.evaluations[~experience.evaluations[match]]
        experience.accuracy = experience.evaluations[match].mean()
        if errors.empty:
            # No errors - nothing to analyze
            experience.errors = errors
            return experience

        # collect errors and create error report
        # first sample errors - make it uniform, but more sophisticated sampling can be implemented
        errors = errors.sample(n=min(3, errors.shape[0]))

        # collect error inputs from runtime
        extra_fields = self._get_extra_fields()
        inputs = student_runtime.process_batch_inputs(
            batch=errors,
            input_template=self.input_template,
            extra_fields=extra_fields
        )

        # construct error report
        errors = pd.concat([
            inputs,
            errors[[self.name, experience.ground_truth_column_name]]
        ], axis=1)
        errors.columns = ['input', 'prediction', 'ground_truth']
        if not teacher_runtime:
            teacher_runtime = student_runtime

        error_reasons = teacher_runtime.process_batch(
            errors,
            instructions="{{#system~}}\n"
                         "LLM prompt was created by concatenating instructions with text input:\n\n"
                         "Prediction = LLM(Input, Instructions)\n\n"
                         "We expect the prediction to be equal to the ground truth.\n"
                         "Your task is to provide a reason for the error due to the original instruction.\n"
                         "Be concise and specific.\n\n"
                         f"Instructions: {self.instructions}\n"
                         "{{~/system}}",
            input_template="{{#user~}}\n"
                           "{{input}}\n"
                           "Prediction: {{prediction}}\n"
                           "Ground truth: {{ground_truth}}\n"
                           "Explanation:\n"
                           "{{~/user}}",
            output_template="{{#assistant~}}{{gen 'reason'}}{{~/assistant}}",
            extra_fields=extra_fields
        )
        errors['reason'] = error_reasons['reason']

        experience.errors = errors
        return experience

    def improve(
        self,
        experience: ShortTermMemory,
        runtime: Runtime,
        update_instructions: bool = True,
    ) -> ShortTermMemory:
        """
        Refines the LLM skill based on its recent experiences.
        
        Args:
            experience (ShortTermMemory): The current experience.
            runtime (Runtime): The runtime instance to be used for processing.
            update_instructions (bool, optional): Flag to decide if instructions should be updated. Defaults to True.
        
        Returns:
            ShortTermMemory: The updated experience after improvements.
        """
        
        experience = experience.model_copy()

        errors = experience.errors.to_dict(orient='records')
        result = runtime.process_record(
            record={
                'errors': errors
            },
            instructions="{{#system~}}\n"
                         "LLM prompt was created by concatenating instructions with text input:\n\n"
                         "Prediction = LLM(Input, Instructions)\n\n"
                         "We expect the prediction to be equal to the ground truth.\n"
                         "Your task is to craft a revised concise instruction for the LLM. "
                         "Follow best practices for LLM prompt engineering.\n"
                         "Include 2-3 examples at the end of your response to demonstrate how the new instruction would be applied.\n"
                         "Use the following format for your examples:\n"
                         "Input: ...\n"
                         "Output: ...\n\n"
                         "{{~/system}}\n",
            input_template="{{#user~}}\n"
                           f"Old instruction: {self.instructions}\n\n"
                           "Errors:\n{{#each errors}}"
                           "\n{{this.input}}\n"
                           "Prediction: {{this.prediction}}\n"
                           "Ground truth: {{this.ground_truth}}\n"
                           "{{/each}}\n"
                           "New instruction:\n"
                           "{{~/user}}",
            output_template="{{#assistant~}}{{gen 'new_instruction'}}{{~/assistant}}",
            extra_fields=self._get_extra_fields()
        )
        new_instruction = result['new_instruction']

        experience.initial_instructions = self.instructions
        experience.updated_instructions = new_instruction

        if update_instructions:
            self.instructions = new_instruction

        return experience



from ..base import LLMSkill
from typing import List


class SequenceLabelingSkill(LLMSkill):
    """
    Skill specialized for sequence labeling on text inputs based on a predefined set of labels.

    This involves tasks like named entity recognition where each word/token in the sequence 
    might be assigned a label.

    Args:
        instructions (str): Templated instruction to guide the LLM in sequence labeling.
        labels (List[str]): A list of valid labels for the sequence labeling task.
        input_template (str): Templated string to format the input for the LLM.
        output_template (str): Templated string to format the output from the LLM.
        prediction_field (str): Specifies the field in which predictions will be stored.

    Note:
        This class is still a work in progress.
    """
    instructions: str = 'Label the input text with the following labels: {{labels}}'
    labels: List[str]
    input_template: str = "Input: {{{{{input}}}}}"
    output_template: str = "Output: {{select 'predictions' options=labels logprobs='score'}}"
    prediction_field: str = 'predictions'



from pydantic import BaseModel, dataclasses, Field, field_validator
from abc import ABC, abstractmethod
from typing import Any, Optional, Dict, Union, Callable

from adala.utils.internal_data import InternalDataFrame, InternalDataFrameConcat
from adala.skills.base import BaseSkill
from adala.memories.base import ShortTermMemory
from adala.datasets import Dataset, DataFrameDataset


class Environment(BaseModel, ABC):
    """Abstract base class for environments.

    The environment provides a mechanism to obtain ground truth information from raw data and predictions, 
    and also facilitates comparison of ground truth with predictions.

    Attributes:
        Config (class): Configuration for the environment class, allows arbitrary types.
    """
        
    @abstractmethod
    def request_feedback(self, skill: BaseSkill, experience: ShortTermMemory):
        """Request user feedback using predictions and update internal ground truth set."""

    @abstractmethod
    def compare_to_ground_truth(self, skill: BaseSkill, experience: ShortTermMemory) -> ShortTermMemory:
        """Compare predictions with ground truth and return the results."""

    @abstractmethod
    def as_dataset(self) -> Dataset:
        """Convert the environment to a dataset."""

    @abstractmethod
    def save(self):
        """Persist the state of the environment."""

    @abstractmethod
    def restore(self):
        """Retrieve and set the state of the environment."""

    class Config:
        arbitrary_types_allowed = True


class BasicEnvironment(Environment):
    """Basic environment implementation.

    This environment assumes the ground truth is provided explicitly with the input data.
    For comparison with ground truth, exact matching is used.

    Attributes:
        ground_truth_dataset (DataFrameDataset): Dataset containing the ground truth data.
                                                 Defaults to an empty DataFrameDataset.
        ground_truth_column (str): Name of the column containing ground truth in the dataset.
                                   Defaults to 'ground_truth'.
        _prediction_column (str): Name of the column containing predictions.

    """
    
    ground_truth_dataset: Union[InternalDataFrame, DataFrameDataset] = Field(default_factory=DataFrameDataset)
    ground_truth_column: str = 'ground_truth'

    _prediction_column: str

    @field_validator('ground_truth_dataset')
    def _validate_ground_truth_dataset(cls, v):
        if isinstance(v, InternalDataFrame):
            return DataFrameDataset(df=v)
        return v

    def request_feedback(self, skill: BaseSkill, experience: ShortTermMemory):
        """In the BasicEnvironment, ground truth is already provided with the input data."""

    def compare_to_ground_truth(self, skill: BaseSkill, experience: ShortTermMemory) -> ShortTermMemory:
        """Compare the predictions with the ground truth using exact matching.

        Args:
            skill (BaseSkill): The skill being evaluated.
            experience (ShortTermMemory): The experience memory containing predictions.

        Returns:
            ShortTermMemory: Updated memory containing evaluation results against ground truth.
        """

        experience = experience.model_copy()

        gt = self.ground_truth_dataset.df[self.ground_truth_column]
        pred = experience.predictions
        # select
        gt = gt[gt.index.isin(pred.index)]
        if gt.empty:
            # return empty memory
            return experience

        gt = gt.to_frame(self.ground_truth_column)

        # compare ground truth with predictions using exact matching
        match_column_name = f'{self.ground_truth_column}__x__{skill.name}'
        evaluations = InternalDataFrameConcat([
            pred,
            (gt[self.ground_truth_column] == pred[skill.name]).rename(match_column_name)
        ], axis=1)
        experience.evaluations = evaluations
        # remember the last column names used in evaluations
        experience.ground_truth_column_name = self.ground_truth_column
        experience.match_column_name = match_column_name
        return experience

    def as_dataset(self) -> Dataset:
        """Return the ground truth dataset.

        Returns:
            Dataset: The dataset containing ground truth data.
        """
        
        return self.ground_truth_dataset

    def save(self):
        """Save method for BasicEnvironment. Not implemented."""
        
        raise NotImplementedError

    def restore(self):
        """Restore method for BasicEnvironment. Not implemented."""
        
        raise NotImplementedError



from .base import TextGenerationSkill


class SummarizationSkill(TextGenerationSkill):
    """
    Skill specialized for summarizing lengthy texts based on the provided input.

    Inherits from the TextGenerationSkill and focuses on generating concise summaries 
    for the input texts. The class customizes the instructions, input, and output templates 
    specifically for text summarization tasks.

    Attributes:
        instructions (str): Instruction to guide the LLM in summarizing the text.
        input_template (str): Format in which the full text is presented to the LLM.
        output_template (str): Expected format of the LLM's summary.
        prediction_field (str): Field name for the generated summary.
    """
    
    instructions: str = 'Summarize the text.'
    input_template: str = "Text: {{{{{input}}}}}"
    output_template: str = "Summary: {{gen 'summary'}}"
    prediction_field: str = 'summary'



import enum
import guidance
import re

from tqdm import tqdm
from abc import ABC, abstractmethod
from pydantic import BaseModel, model_validator
from typing import List, Dict, Optional, Tuple, Any
from adala.datasets.base import InternalDataFrame

tqdm.pandas()


class Runtime(BaseModel, ABC):
    """
    Base class representing a generic runtime environment.

    Attributes:
        verbose (bool): Flag indicating if runtime outputs should be verbose. Defaults to False.
    """
    verbose: bool = False

    @model_validator(mode='after')
    def init_runtime(self):
        """Initializes the runtime.

        This method should be used to validate and potentially initialize the runtime instance.

        Returns:
            Runtime: The initialized runtime instance.
        """  
        return self


class LLMRuntimeType(enum.Enum):
    STUDENT = 'student'
    TEACHER = 'teacher'


class LLMRuntimeModelType(enum.Enum):
    """Enumeration for LLM runtime model types."""    
    OpenAI = 'OpenAI'
    Transformers = 'Transformers'


class LLMRuntime(Runtime):
    """
    Class representing an LLM runtime environment.

    Attributes:
        llm_runtime_type (LLMRuntimeModelType): Type of the LLM runtime. Defaults to OpenAI.
        llm_params (Dict[str, str]): Parameters for the LLM runtime. Defaults to a basic GPT-3.5 configuration.
    
        _llm: Internal instance for the LLM model. Initialized in `init_runtime`.
        _program: Program instance used for guidance. Initialized in `init_runtime`.
        _llm_template (str): Template string for LLM guidance.
    """
    llm_runtime_type: LLMRuntimeType = LLMRuntimeType.STUDENT
    llm_runtime_model_type: LLMRuntimeModelType = LLMRuntimeModelType.OpenAI
    llm_params: Dict[str, str] = {
        'model': 'gpt-3.5-turbo-instruct',
        # 'max_tokens': 10,
        # 'temperature': 0,
    }
    _llm = None
    _program = None
    # do not override this template
    _llm_template: str = '''\
{{>instructions_program}}

{{>input_program}}
{{>output_program}}'''

    class Config:
        arbitrary_types_allowed = True

    def _create_program(self):
        # create an LLM instance
        if self.llm_runtime_model_type.value == LLMRuntimeModelType.OpenAI.value:
            self._llm = guidance.llms.OpenAI(**self.llm_params)
        elif self.llm_runtime_model_type.value == LLMRuntimeModelType.Transformers.value:
            self._llm = guidance.llms.Transformers(**self.llm_params)
        else:
            raise NotImplementedError(f'LLM runtime type {self.llm_runtime_model_type} is not implemented.')
        self._program = guidance(self._llm_template, llm=self._llm, silent=not self.verbose)

    def init_runtime(self):
        """Initializes the LLM runtime environment.

        Creates an LLM instance based on the runtime type and parameters.

        Returns:
            LLMRuntime: Initialized runtime instance.
        """
        self._create_program()
        return self

    def get_outputs(self, output_template: str) -> List[str]:
        """Extracts output fields from the output template.

        Args:
            output_template (str): The template string to extract output fields from.

        Returns:
            List[str]: List of extracted output fields.
        """
        # search for all occurrences of {{...'output'...}}
        # TODO: this is a very naive regex implementation - likely to fail in many cases
        outputs = re.findall(r'\'(.*?)\'', output_template)
        return outputs

    def _process_record(
        self,
        record,
        program,
        extra_fields,
        outputs=None
    ):
        """Processes a single record using the guidance program.

        Args:
            record (dict or InternalDataFrame): The record to be processed.
            program (callable): The guidance program for processing.
            extra_fields (dict, optional): Additional fields to include in the processed record.
            outputs (list of str, optional): Specific output fields to extract from the result.

        Returns:
            dict: Processed output for the record.
        """
        
        if not isinstance(record, dict):
            record = record.to_dict()
        else:
            record = record.copy()
        verified_input = record
        # exclude guidance parameter from input
        if 'text' in verified_input:
            verified_input['text_'] = verified_input['text']
            del verified_input['text']

        verified_input.update(extra_fields)
        result = program(
            silent=not self.verbose,
            **verified_input
        )
        if outputs is None:
            verified_output = {'': str(result)}
        else:
            verified_output = {field: result[field] for field in outputs}

        return verified_output

    def get_input_program(self, input_template):
        """Generates an input program from the provided template.

        Args:
            input_template (str): Template to generate the input program.

        Returns:
            callable: The generated input program.
        """
        
        # fix input template in case "text" is presented there - there might be other paramater names as well...
        fixed_input_template = input_template
        if '{{text}}' in fixed_input_template:
            fixed_input_template = fixed_input_template.replace('{{text}}', '{{text_}}')
        input_program = guidance(fixed_input_template, llm=self._llm, silent=not self.verbose)
        return input_program

    def get_output_program(self, output_template):
        """Generates an output program from the provided template.

        Args:
            output_template (str): Template to generate the output program.

        Returns:
            callable: The generated output program.
        """
        
        return guidance(output_template, llm=self._llm)

    def get_instructions_program(self, instructions):
        """Generates an instructions program from the provided template.

        Args:
            instructions (str): The instructions to generate the program.

        Returns:
            callable: The generated instructions program.
        """
        
        return guidance(instructions, llm=self._llm)

    def process_record(
        self,
        record: Dict[str, Any],
        input_template: str,
        output_template: str,
        instructions: str,
        extra_fields: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Processes a record using the provided templates and instructions.

        Args:
            record (Dict[str, Any]): The record data to be processed.
            input_template (str): Template for input processing.
            output_template (str): Template for output processing.
            instructions (str): Instructions for guidance.
            extra_fields (Dict[str, Any], optional): Additional fields to include during processing.

        Returns:
            Dict[str, Any]: The processed record.
        """
        
        outputs = re.findall(r'\'(.*?)\'', output_template)

        input = record.copy()
        input.update({
            'input_program': self.get_input_program(input_template),
            'output_program': self.get_output_program(output_template),
            'instructions_program': self.get_instructions_program(instructions),
        })
        output = self._process_record(
            record=input,
            program=self._program,
            outputs=outputs,
            extra_fields=extra_fields
        )
        return output

    def process_batch(
        self,
        batch: InternalDataFrame,
        input_template: str,
        output_template: str,
        instructions: str,
        extra_fields: Optional[Dict[str, Any]] = None,
    ) -> InternalDataFrame:
        """Processes a batch of records using the provided templates and instructions.

        Args:
            batch (InternalDataFrame): The batch of records to be processed.
            input_template (str): Template for input processing.
            output_template (str): Template for output processing.
            instructions (str): Instructions for guidance.
            extra_fields (Dict[str, Any], optional): Additional fields to include during batch processing.

        Returns:
            InternalDataFrame: The processed batch of records.
        """
        
        outputs = self.get_outputs(output_template)

        extra_fields = extra_fields or {}
        # copy extra fields to avoid modification of the original dict
        extra_fields = extra_fields.copy()
        # TODO: it's not efficient way to initialize the program here - should be done once
        extra_fields.update({
            'input_program': self.get_input_program(input_template),
            'output_program': self.get_output_program(output_template),
            'instructions_program': self.get_instructions_program(instructions),
        })
        output = batch.progress_apply(
            self._process_record,
            axis=1,
            result_type='expand',
            program=self._program,
            outputs=outputs,
            extra_fields=extra_fields
        )
        return output

    def process_batch_inputs(
        self,
        batch: InternalDataFrame,
        input_template: str,
        extra_fields: Optional[Dict[str, Any]] = None,
    ) -> InternalDataFrame:
        """Processes inputs for a batch of records using the provided input template.

        Args:
            batch (InternalDataFrame): The batch of records for input processing.
            input_template (str): The template for input processing.
            extra_fields (Dict[str, Any], optional): Additional fields to include during input processing.

        Returns:
            InternalDataFrame: The processed inputs for the batch of records.
        """
        
        output = batch.progress_apply(
            self._process_record,
            axis=1,
            result_type='expand',
            program=self.get_input_program(input_template),
            extra_fields=extra_fields or {}
        )
        return output


class CodeRuntime(Runtime):
    """Base class representing a runtime designed for executing code."""



from pydantic import BaseModel, model_validator
from abc import ABC, abstractmethod
from typing import List, Union, Dict, Any, Optional
from adala.datasets.base import Dataset
from adala.runtimes.base import Runtime
from adala.memories.base import ShortTermMemory
from .base import BaseSkill


class SkillSet(BaseModel, ABC):
    """
    Represents a collection of interdependent skills aiming to achieve a specific goal.
    
    A skill set breaks down the path to achieve a goal into necessary precursor skills.
    Agents can evolve these skills either in parallel for tasks like self-consistency or 
    sequentially for complex problem decompositions and causal reasoning. In the most generic
    cases, task decomposition can involve a graph-based approach.

    Args:
        skills (Dict[str, BaseSkill]): Dictionary mapping skill names to their corresponding 
                                       BaseSkill instances.
    """
    
    skills: Dict[str, BaseSkill]

    @abstractmethod
    def apply(self, dataset: Dataset, runtime: Runtime, experience: Optional[ShortTermMemory] = None) -> ShortTermMemory:
        """
        Apply the skill set to a dataset using a specified runtime.
        
        Args:
            dataset (Dataset): The dataset to apply the skill set to.
            runtime (Runtime): The runtime environment in which to apply the skills.
            experience (Optional[ShortTermMemory], optional): Existing experience data. Defaults to None.
            
        Returns:
            ShortTermMemory: Updated experience after applying the skill set.
        """

    @abstractmethod
    def select_skill_to_improve(self, experience: ShortTermMemory) -> BaseSkill:
        """
        Select the next skill to enhance based on the current experience.
        
        Args:
            experience (ShortTermMemory): Current experience data.
            
        Returns:
            BaseSkill: Skill selected for improvement.
        """


class LinearSkillSet(SkillSet):
    """
    Represents a sequence of skills that are acquired in a specific order to achieve a goal.

    LinearSkillSet ensures that skills are developed in a sequential manner, determined either 
    by the provided skill_sequence or by the lexicographical order of skill names.

    Args:
        skill_sequence (List[str], optional): Ordered list of skill names indicating the order 
                                              in which they should be acquired.
    """
    
    skill_sequence: List[str] = None

    @model_validator(mode='after')
    def skill_sequence_validator(self):
        """
        Validates and sets the default order for the skill sequence if not provided.
        
        Returns:
            LinearSkillSet: The current instance with updated skill_sequence attribute.
        """
        
        if self.skill_sequence is None:
            # use default skill sequence defined by lexicographical order
            self.skill_sequence = sorted(self.skills.keys())
        return self

    def apply(
        self, dataset: Dataset,
        runtime: Runtime,
        experience: Optional[ShortTermMemory] = None
    ) -> ShortTermMemory:
        """
        Sequentially applies each skill on the dataset, enhancing the agent's experience.
        
        Args:
            dataset (Dataset): The dataset to apply the skills on.
            runtime (Runtime): The runtime environment in which to apply the skills.
            experience (Optional[ShortTermMemory], optional): Existing experience data. Defaults to None.
            
        Returns:
            ShortTermMemory: Updated experience after sequentially applying the skills.
        """
        if experience is None:
            experience = ShortTermMemory(dataset=dataset)
        else:
            experience = experience.model_copy()

        for skill_name in self.skill_sequence:
            skill = self.skills[skill_name]
            experience = skill.apply(dataset, runtime, experience)
        
        return experience

    def select_skill_to_improve(self, experience: ShortTermMemory) -> BaseSkill:
        """
        Picks the next skill for improvement in the sequence.
        
        Args:
            experience (ShortTermMemory): Current experience data.
            
        Returns:
            BaseSkill: The next skill selected for improvement.
        """
        
        # TODO: implement real logic for skill selection
        return self.skills[self.skill_sequence[0]]

    def __rich__(self):
        """Returns a rich representation of the skill."""
        # TODO: move it to a base class and use repr derived from Skills
        text = f"[bold blue]Total Agent Skills: {len(self.skills)}[/bold blue]\n\n"
        for skill in self.skills.values():
            text += f'[bold underline green]{skill.name}[/bold underline green]\n' \
                    f'[green]{skill.instructions}[green]\n'
        return text


class ParallelSkillSet(SkillSet):
    """
    Represents a set of skills that are acquired simultaneously to reach a goal.
    """
    
    pass



from .base import TextGenerationSkill


class TranslationSkill(TextGenerationSkill):
    """
    Skill specialized for translating text from one language to another.

    Inherits from the TextGenerationSkill and focuses on translating the input text to the
    specified target language. The class customizes the instructions, input, and output templates
    specifically for translation tasks.

    Attributes:
        instructions (str): Instruction to guide the LLM in translating the text.
        input_template (str): Format in which the full text is presented to the LLM.
        output_template (str): Expected format of the LLM's translation.
        prediction_field (str): Field name for the generated translation.
        target_language (str): Language to which the input text is translated.
    """

    name: str = 'translation'
    description: str = 'Translate text from one language to another.'
    instructions: str = 'Identify the language of the given text and translate it to {{target_language}}.'
    input_template: str = "Text: {{{{{input}}}}}"
    # output_template: str = "Input language: {{gen 'detected_language'}}\nTranslation: {{gen 'translation'}}"
    output_template: str = "Translation: {{gen 'translation'}}"
    prediction_field: str = 'translation'
    target_language: str = 'English'



from .base import LongTermMemory, ShortTermMemory
from typing import Any


class FileMemory(LongTermMemory):

    filepath: str

    def remember(self, experience: ShortTermMemory):
        """
        Serialize experience in JSON and append to file
        """
        experience_json = experience.model_dump_json()
        with open(self.filepath, 'a') as f:
            f.write(experience_json + '\n')

    def retrieve(self, observations: ShortTermMemory) -> ShortTermMemory:
        """
        Retrieve experience from file
        """
        raise NotImplementedError



import pandas as pd
from unittest.mock import MagicMock, patch
from adala.runtimes.openai import OpenAIRuntime

from adala.agents import Agent
from adala.datasets import DataFrameDataset
from adala.environments import BasicEnvironment
from adala.skills import ClassificationSkill
from adala.utils.logs import print_dataframe


def process_record_generator(*args, **kwargs):
    # train
    for i in range(3):
        # predictions for gt comparison
        yield {'sentiment': 'Neutral' if i < 2 else 'Positive'}
        yield {'sentiment': 'Neutral' if i < 2 else 'Negative'}
        yield {'sentiment': 'Neutral'}

        # errors
        yield {'reason': 'Test reason'}
        yield {'reason': 'Test reason'}
        yield {'reason': 'Test reason'}
        yield {'reason': 'Test reason'}

        # instruction generation
        yield {'new_instruction': 'Test instruction'}

    # test
    yield {'sentiment': 'Positive'}
    yield {'sentiment': 'Negative'}
    yield {'sentiment': 'Neutral'}


@patch.object(OpenAIRuntime, '_check_api_key', return_value=None)
@patch.object(OpenAIRuntime, '_check_model_availability', return_value=None)
@patch.object(OpenAIRuntime, '_process_record', side_effect=process_record_generator())
def test_classification_skill(
        mock_check_api_key,
        mock_check_model_availability,
        mock_process_record
):
    print("=> Initialize datasets ...")

    # Train dataset
    train_df = pd.DataFrame([
        ["It was the negative first impressions, and then it started working.", "Positive"],
        ["Not loud enough and doesn't turn on like it should.", "Negative"],
        ["I don't know what to say.", "Neutral"],
    ], columns=["text", "ground_truth"])

    # Test dataset
    test_df = pd.DataFrame([
        "All three broke within two months of use.",
        "The device worked for a long time, can't say anything bad.",
        "Just a random line of text.",
    ], columns=["text"])

    train_dataset = DataFrameDataset(df=train_df)
    test_dataset = DataFrameDataset(df=test_df)

    print("=> Initialize and train ADALA agent ...")
    agent = Agent(
        # connect to a dataset
        environment=BasicEnvironment(
            ground_truth_dataset=train_dataset,
            ground_truth_column="ground_truth"
        ),
        # define a skill
        skills=ClassificationSkill(
            name='sentiment',
            instructions="Label text as subjective or objective.",
            labels=["Positive", "Negative", "Neutral"],
            input_data_field='text'
        ),
    )
    run = agent.learn(learning_iterations=3, accuracy_threshold=0.95)
    assert run.accuracy > 0.8

    print('\n\n=> Final instructions:')
    print('=====================')
    print(f'{run.updated_instructions}')
    print('=====================')

    print('\n=> Run test ...')
    run = agent.apply_skills(test_dataset)
    print_dataframe(run.predictions)

    assert not run.predictions.empty



from ..base import LLMSkill
from typing import List


class ClassificationSkill(LLMSkill):
    """
    Skill specialized for classifying text inputs based on a predefined set of labels.

    Args:
        instructions (str): Templated instruction to guide the LLM in classification.
        labels (List[str]): A list of valid labels for the classification task.
        output_template (str): Templated string to format the output from the LLM.
        prediction_field (str): Specifies the field in which predictions will be stored.    
    """
    
    instructions: str = 'Label the input text with the following labels: {{labels}}'
    labels: List[str]
    output_template: str = "Output: {{select 'predictions' options=labels logprobs='score'}}"
    prediction_field: str = 'predictions'


class ClassificationSkillWithCoT(ClassificationSkill):
    """
    Skill specialized for classifying text inputs with the addition of generating a Chain of Thought.

    Args:
        instructions (str): Templated instruction to guide the LLM in classification and to generate a rationale.
        labels (List[str]): A list of valid labels for the classification task.
        input_template (str): Templated string to format the input, which includes a rationale (thoughts).
        output_template (str): Templated string to format the output from the LLM.
        prediction_field (str): Specifies the field in which predictions will be stored.
    """
    
    instructions: str = 'Label the input text with the following labels: {{labels}}. Provide a rationale for your answer.'
    output_template: str = "Thoughts: {{gen 'rationale'}}\nOutput: {{select 'predictions' options=labels logprobs='score'}}"



import os
import openai
from pydantic import model_validator, field_validator, ValidationError, ValidationInfo, Field
from typing import Optional, Dict
from .base import LLMRuntime, LLMRuntimeType, LLMRuntimeModelType
from adala.utils.logs import print_error


class OpenAIRuntime(LLMRuntime):
    """Runtime class specifically designed for OpenAI models.

    This class is tailored to use OpenAI models, particularly GPT models.
    It inherits from the `LLMRuntime` class and thus can utilize its functionalities but specializes 
    for the OpenAI ecosystem.

    Attributes:
        api_key (str): The API key required to access OpenAI's API.
        gpt_model_name (str): Name of the GPT model. Defaults to 'gpt-3.5-turbo-instruct'.
        temperature (float): Sampling temperature for the GPT model's output. 
                             A higher value makes output more random, while a lower value makes it more deterministic.
                             Defaults to 0.0.
    """
    
    api_key: Optional[str] = None
    gpt_model_name: Optional[str] = Field(default='gpt-3.5-turbo-instruct', alias='model')
    temperature: Optional[float] = 0.0

    def _check_api_key(self):
        if self.api_key:
            return
        self.api_key = os.getenv('OPENAI_API_KEY')
        if not self.api_key:
            print_error(
                'OpenAI API key is not provided. Please set the OPENAI_API_KEY environment variable:\n\n'
                'export OPENAI_API_KEY=your-openai-api-key\n\n'
                'or set the `api_key` attribute of the `OpenAIRuntime` python class:\n\n'
                f'{self.__class__.__name__}(..., api_key="your-openai-api-key")\n\n'
                f'Read more about OpenAI API keys at https://platform.openai.com/docs/quickstart/step-2-setup-your-api-key')
            raise ValidationError('OpenAI API key is not provided.')

    def _check_model_availability(self):
        models = openai.Model.list(api_key=self.api_key)
        models = set(model['id'] for model in models['data'])
        if self.gpt_model_name not in models:
            print_error(
                f'Requested model "{self.gpt_model_name}" is not available in your OpenAI account. '
                f'Available models are: {models}\n\n'
                f'Try to change the runtime settings for {self.__class__.__name__}, for example:\n\n'
                f'{self.__class__.__name__}(..., model="gpt-3.5-turbo")\n\n'
            )
            raise ValidationError(f'Requested model {self.gpt_model_name} is not available in your OpenAI account.')

    def init_runtime(self):
        self._check_api_key()
        self._check_model_availability()

        student_models = {'gpt-3.5-turbo-instruct', 'text-davinci-003'}
        teacher_models = {'gpt-4', 'gpt-3.5-turbo', 'gpt-3.5-turbo-16k'}

        if self.gpt_model_name in student_models:
            self.llm_runtime_type = LLMRuntimeType.STUDENT
        elif self.gpt_model_name in teacher_models:
            self.llm_runtime_type = LLMRuntimeType.TEACHER
        else:
            raise NotImplementedError(f'Not supported model: {self.gpt_model_name}.')

        self.llm_runtime_model_type = LLMRuntimeModelType.OpenAI
        self.llm_params = {
            'model': self.gpt_model_name,
            'temperature': self.temperature,
            'api_key': self.api_key
        }
        self._create_program()
        return self



import pandas as pd
import time

from rich import print
from rich.table import Table
from rich import box
from rich.console import Console
from typing import Optional
from .internal_data import InternalDataFrame

console = Console()
error_console = Console(stderr=True, style="bold red")


def print_text(text: str, style=None, streaming_style=False):
    if streaming_style:
        for char in text:
            console.print(char, sep='', end='', style=style)
            time.sleep(0.01)
        console.print()
    else:
        console.print(text, style=style)


def print_error(text: str):
    error_console.print(text)


def print_dataframe(dataframe: InternalDataFrame):
    num_rows = 5
    table = Table(show_header=True, header_style="bold magenta")
    # index_name = dataframe.index.name or 'index'
    # table.add_column(index_name)

    for column in dataframe.columns:
        table.add_column(str(column))

    for index, value_list in enumerate(dataframe.iloc[:num_rows].values.tolist()):
        # row = [str(index)]
        row = []
        row += [str(x) for x in value_list]
        table.add_row(*row)

    # Update the style of the table
    table.row_styles = ["none", "dim"]
    table.box = box.SIMPLE_HEAD

    console.print(table)



from .labeling.classification import LLMSkill, ClassificationSkill, ClassificationSkillWithCoT
from .labeling.sequence_labeling import SequenceLabelingSkill
from .generation.base import TextGenerationSkill
from .generation.qa import QuestionAnsweringSkill
from .generation.summarization import SummarizationSkill




import json
import label_studio_sdk

from pprint import pprint
from .base import Dataset, InternalDataFrame
from pydantic import model_validator, SkipValidation
from label_studio_sdk.project import LabelStudioException, Project
from typing import Optional, List, Dict


class LabelStudioFormatMixin:

    def _tasks_to_df(
        self,
        tasks,
        include_annotations: bool = True,
        only_annotated: bool = False,
        ground_truth_column: str = 'ground_truth'
    ):
        indices, records = [], []
        for task in tasks:
            record = task['data']
            if only_annotated and not task['annotations']:
                continue

            if (only_annotated or include_annotations) and task['annotations']:
                # TODO: expand more complex annotations
                if len(task['annotations']) > 1:
                    raise NotImplementedError('Multiple annotations are not supported yet')
                annotation = task['annotations'][0]
                annotation_type = annotation['result'][0]['type']
                if annotation_type == 'textarea':
                    annotation_type = 'text'
                if len(annotation['result']) > 1:
                    raise NotImplementedError('Multiple results per annotation are not supported yet')
                label = annotation['result'][0]['value'][annotation_type]
                if isinstance(label, list):
                    if len(label) == 1:
                        label = label[0]
                    else:
                        label = ','.join(sorted(label))
                else:
                    label = str(label)
                record[ground_truth_column] = label

            index = task['id']
            records.append(record)
            indices.append(index)
        return InternalDataFrame(records, index=indices)


class LabelStudioDataset(Dataset, LabelStudioFormatMixin):

    label_studio_url: str
    label_studio_api_key: str
    label_studio_project_id: int

    ground_truth_column: str = 'ground_truth'

    _project_client: SkipValidation[Project] = None

    @model_validator(mode='after')
    def init_client(self):
        if self._project_client is None:
            client = label_studio_sdk.Client(
                url=self.label_studio_url,
                api_key=self.label_studio_api_key
            )
            self._project_client = client.get_project(id=self.label_studio_project_id)
        return self

    def get_project_info(self):
        return self._project_client.get_params()

    def info(self) -> None:
        pprint(self.get_project_info())

    def __len__(self):
        info = self.get_project_info()
        return info['task_number']

    def batch_iterator(self, batch_size: int = 100) -> InternalDataFrame:
        page = 1
        while True:
            try:
                data = self._project_client.get_paginated_tasks(page=page, page_size=batch_size)
                yield self._tasks_to_df(data['tasks'], include_annotations=False)
                page += 1
            # we'll get 404 from API on empty page
            except LabelStudioException as e:
                break

    def get_ground_truth(self, batch: Optional[InternalDataFrame] = None) -> InternalDataFrame:
        if batch is None:
            labeled_tasks = self._project_client.get_labeled_tasks()
            gt = self._tasks_to_df(labeled_tasks, only_annotated=True, ground_truth_column='ground_truth')
            return gt
        else:
            # TODO: not the most effective method - better to send subset of indices to LS API
            labeled_tasks = self._project_client.get_labeled_tasks()
            gt = self._tasks_to_df(labeled_tasks, only_annotated=True, ground_truth_column='ground_truth')
            return gt[gt.index.isin(batch.index)]


class LabelStudioFileDataset(Dataset, LabelStudioFormatMixin):
    label_studio_file: str
    ground_truth_column: str = 'ground_truth'

    _data: List[Dict] = None

    @model_validator(mode='after')
    def load_data(self):
        with open(self.label_studio_file) as f:
            self._data = json.load(f)
        return self

    def batch_iterator(self, batch_size: int = 100) -> InternalDataFrame:
        for i in range(0, len(self._data), batch_size):
            batch = self._data[i:i+batch_size]
            yield self._tasks_to_df(batch, include_annotations=False)

    def get_ground_truth(self, batch: Optional[InternalDataFrame]) -> InternalDataFrame:
        return self._tasks_to_df(self._data, only_annotated=True, ground_truth_column='ground_truth')

    def __len__(self):
        return len(self._data)

    def info(self) -> None:
        print(f'Total Label Studio tasks loaded: {len(self)}')



from typing import Iterable
from .base import Dataset
from adala.utils.internal_data import InternalDataFrame
from pydantic import Field


class DataFrameDataset(Dataset):
    """
    Represents a dataset backed by an internal data frame.

    Provides methods to interact with and obtain information about the dataset stored
    as an internal data frame. This class wraps around `InternalDataFrame` to make it 
    compatible with the dataset abstraction.

    Attributes:
        df (InternalDataFrame): The internal data frame storing the dataset.
    """
    
    df: InternalDataFrame = Field(default_factory=InternalDataFrame)

    class Config:
        arbitrary_types_allowed = True

    def __len__(self):
        """
        Provides the number of records in the dataset.
        
        Returns:
            int: Total number of records in the dataset.
        """
        
        return len(self.df)

    def batch_iterator(self, batch_size: int = 100) -> Iterable[InternalDataFrame]:
        """
        Yields batches of data records from the dataset.
        
        Args:
            batch_size (int, optional): Size of each batch to be yielded. Defaults to 100.
            
        Yields:
            Iterable[InternalDataFrame]: An iterator that yields data frames containing batches of records.
        """
        
        for i in range(0, len(self.df), batch_size):
            yield self.df.iloc[i:i+batch_size]

    def info(self) -> None:
        """
        Displays information (statistical description) about the dataset.
        """
        
        print(self.df.describe())



import pandas as pd
from typing import List, Dict, Any, Union, Iterable

RawRecord = Dict[str, Any]
RawRecords = List[RawRecord]

# Internal data tables representation. Replace this with Dask or Polars in the future.
InternalDataFrame = pd.DataFrame


def InternalDataFrame_encoder(df: InternalDataFrame) -> List:
    return df.to_dict(orient='records')


def InternalDataFrameConcat(dfs: Iterable[InternalDataFrame], **kwargs) -> InternalDataFrame:
    return pd.concat(dfs, **kwargs)
